import os
import pathlib

import google.auth
from ibis.backends.tests.base import BackendTest, RoundAwayFromZero, UnorderedComparator

import ibis_bigquery

DATASET_ID = "ibis_gbq_testing"
DEFAULT_PROJECT_ID = "ibis-gbq"
PROJECT_ID_ENV_VAR = "GOOGLE_BIGQUERY_PROJECT_ID"


class TestConf(UnorderedComparator, BackendTest, RoundAwayFromZero):
    """Backend-specific class with information for testing."""

    # These were moved from TestConf for use in common test suite.
    # TODO: Indicate RoundAwayFromZero and UnorderedComparator.
    # https://github.com/ibis-project/ibis-bigquery/issues/30
    supports_divide_by_zero = True
    supports_floating_modulus = False
    returned_timestamp_unit = "us"

    def name(self):
        """Name of the backend.
        In the parent class, this is automatically obtained from the name of
        the module, which is not the case for third-party backends.
        """
        return "bigquery"

    @staticmethod
    def connect(data_directory: pathlib.Path) -> ibis_bigquery.Backend:
        """Connect to the test database."""
        credentials, default_project_id = google.auth.default(
            scopes=ibis_bigquery.EXTERNAL_DATA_SCOPES
        )

        project_id = os.getenv(PROJECT_ID_ENV_VAR)
        if project_id is None:
            project_id = default_project_id
        if project_id is None:
            project_id = DEFAULT_PROJECT_ID

        return ibis_bigquery.connect(
            project_id=project_id,
            dataset_id=f"{project_id}.{DATASET_ID}",
            credentials=credentials,
        )



import datetime

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import packaging.version
import pandas as pd
import pytest
from ibis.expr.types import TableExpr

import ibis_bigquery

IBIS_VERSION = packaging.version.Version(ibis.__version__)
IBIS_1_4_VERSION = packaging.version.Version("1.4.0")


@pytest.mark.parametrize(
    ("case", "expected", "dtype"),
    [
        (datetime.date(2017, 1, 1), "DATE '2017-01-01'", dt.date),
        (pd.Timestamp("2017-01-01"), "DATE '2017-01-01'", dt.date,),
        ("2017-01-01", "DATE '2017-01-01'", dt.date),
        (
            datetime.datetime(2017, 1, 1, 4, 55, 59),
            "TIMESTAMP '2017-01-01 04:55:59'",
            dt.timestamp,
        ),
        ("2017-01-01 04:55:59", "TIMESTAMP '2017-01-01 04:55:59'", dt.timestamp,),
        (
            pd.Timestamp("2017-01-01 04:55:59"),
            "TIMESTAMP '2017-01-01 04:55:59'",
            dt.timestamp,
        ),
    ],
)
def test_literal_date(case, expected, dtype):
    expr = ibis.literal(case, type=dtype).year()
    result = ibis_bigquery.compile(expr)
    assert result == f"SELECT EXTRACT(year from {expected}) AS `tmp`"


@pytest.mark.parametrize(
    ("case", "expected", "dtype", "strftime_func"),
    [
        (datetime.date(2017, 1, 1), "DATE '2017-01-01'", dt.date, "FORMAT_DATE",),
        (pd.Timestamp("2017-01-01"), "DATE '2017-01-01'", dt.date, "FORMAT_DATE",),
        ("2017-01-01", "DATE '2017-01-01'", dt.date, "FORMAT_DATE",),
        (
            datetime.datetime(2017, 1, 1, 4, 55, 59),
            "TIMESTAMP '2017-01-01 04:55:59'",
            dt.timestamp,
            "FORMAT_TIMESTAMP",
        ),
        (
            "2017-01-01 04:55:59",
            "TIMESTAMP '2017-01-01 04:55:59'",
            dt.timestamp,
            "FORMAT_TIMESTAMP",
        ),
        (
            pd.Timestamp("2017-01-01 04:55:59"),
            "TIMESTAMP '2017-01-01 04:55:59'",
            dt.timestamp,
            "FORMAT_TIMESTAMP",
        ),
    ],
)
def test_day_of_week(case, expected, dtype, strftime_func):
    date_var = ibis.literal(case, type=dtype)
    expr_index = date_var.day_of_week.index()
    result = ibis_bigquery.compile(expr_index)
    assert result == f"SELECT MOD(EXTRACT(DAYOFWEEK FROM {expected}) + 5, 7) AS `tmp`"

    expr_name = date_var.day_of_week.full_name()
    result = ibis_bigquery.compile(expr_name)
    if strftime_func == "FORMAT_TIMESTAMP":
        assert result == f"SELECT {strftime_func}('%A', {expected}, 'UTC') AS `tmp`"
    else:
        assert result == f"SELECT {strftime_func}('%A', {expected}) AS `tmp`"


@pytest.mark.parametrize(
    ("case", "expected", "dtype"),
    [
        ("test of hash", "'test of hash'", dt.string,),
        (b"test of hash", "FROM_BASE64('dGVzdCBvZiBoYXNo')", dt.binary,),
    ],
)
def test_hash(case, expected, dtype):
    if IBIS_VERSION < IBIS_1_4_VERSION:
        pytest.skip("requires ibis 1.4+")
    string_var = ibis.literal(case, type=dtype)
    expr = string_var.hash(how="farm_fingerprint")
    result = ibis_bigquery.compile(expr)
    assert result == f"SELECT farm_fingerprint({expected}) AS `tmp`"


@pytest.mark.parametrize(
    ("case", "expected", "how", "dtype"),
    [
        ("test", "md5('test')", "md5", dt.string,),
        (b"test", "md5(FROM_BASE64('dGVzdA=='))", "md5", dt.binary,),
        ("test", "sha1('test')", "sha1", dt.string,),
        (b"test", "sha1(FROM_BASE64('dGVzdA=='))", "sha1", dt.binary,),
        ("test", "sha256('test')", "sha256", dt.string,),
        (b"test", "sha256(FROM_BASE64('dGVzdA=='))", "sha256", dt.binary,),
        ("test", "sha512('test')", "sha512", dt.string,),
        (b"test", "sha512(FROM_BASE64('dGVzdA=='))", "sha512", dt.binary,),
    ],
)
def test_hashbytes(case, expected, how, dtype):
    if IBIS_VERSION < IBIS_1_4_VERSION:
        pytest.skip("requires ibis 1.4+")
    var = ibis.literal(case, type=dtype)
    expr = var.hashbytes(how=how)
    result = ibis_bigquery.compile(expr)
    assert result == f"SELECT {expected} AS `tmp`"


@pytest.mark.parametrize(
    ("case", "expected", "dtype"),
    [
        (
            datetime.datetime(2017, 1, 1, 4, 55, 59),
            "TIMESTAMP '2017-01-01 04:55:59'",
            dt.timestamp,
        ),
        ("2017-01-01 04:55:59", "TIMESTAMP '2017-01-01 04:55:59'", dt.timestamp,),
        (
            pd.Timestamp("2017-01-01 04:55:59"),
            "TIMESTAMP '2017-01-01 04:55:59'",
            dt.timestamp,
        ),
        (datetime.time(4, 55, 59), "TIME '04:55:59'", dt.time),
        ("04:55:59", "TIME '04:55:59'", dt.time),
    ],
)
def test_literal_timestamp_or_time(case, expected, dtype):
    expr = ibis.literal(case, type=dtype).hour()
    result = ibis_bigquery.compile(expr)
    assert result == f"SELECT EXTRACT(hour from {expected}) AS `tmp`"


def test_projection_fusion_only_peeks_at_immediate_parent():
    if IBIS_VERSION < IBIS_1_4_VERSION:
        pytest.skip("requires ibis 1.4+")
    schema = [
        ("file_date", "timestamp"),
        ("PARTITIONTIME", "date"),
        ("val", "int64"),
    ]
    table = ibis.table(schema, name="unbound_table")
    table = table[table.PARTITIONTIME < ibis.date("2017-01-01")]
    table = table.mutate(file_date=table.file_date.cast("date"))
    table = table[table.file_date < ibis.date("2017-01-01")]
    table = table.mutate(XYZ=table.val * 2)
    expr = table.join(table.view())[table]
    result = ibis_bigquery.compile(expr)
    expected = """\
WITH t0 AS (
  SELECT *
  FROM unbound_table
  WHERE `PARTITIONTIME` < DATE '2017-01-01'
),
t1 AS (
  SELECT CAST(`file_date` AS DATE) AS `file_date`, `PARTITIONTIME`, `val`
  FROM t0
),
t2 AS (
  SELECT t1.*
  FROM t1
  WHERE t1.`file_date` < DATE '2017-01-01'
),
t3 AS (
  SELECT *, `val` * 2 AS `XYZ`
  FROM t2
)
SELECT t3.*
FROM t3
  INNER JOIN t3 t4"""
    assert result == expected


@pytest.mark.parametrize(
    ("unit", "expected_unit", "expected_func"),
    [
        ("Y", "YEAR", "TIMESTAMP"),
        ("Q", "QUARTER", "TIMESTAMP"),
        ("M", "MONTH", "TIMESTAMP"),
        ("W", "WEEK", "TIMESTAMP"),
        ("D", "DAY", "TIMESTAMP"),
        ("h", "HOUR", "TIMESTAMP"),
        ("m", "MINUTE", "TIMESTAMP"),
        ("s", "SECOND", "TIMESTAMP"),
        ("ms", "MILLISECOND", "TIMESTAMP"),
        ("us", "MICROSECOND", "TIMESTAMP"),
        ("Y", "YEAR", "DATE"),
        ("Q", "QUARTER", "DATE"),
        ("M", "MONTH", "DATE"),
        ("W", "WEEK", "DATE"),
        ("D", "DAY", "DATE"),
        ("h", "HOUR", "TIME"),
        ("m", "MINUTE", "TIME"),
        ("s", "SECOND", "TIME"),
        ("ms", "MILLISECOND", "TIME"),
        ("us", "MICROSECOND", "TIME"),
    ],
)
def test_temporal_truncate(unit, expected_unit, expected_func):
    t = ibis.table([("a", getattr(dt, expected_func.lower()))], name="t")
    expr = t.a.truncate(unit)
    result = ibis_bigquery.compile(expr)
    expected = f"""\
SELECT {expected_func}_TRUNC(`a`, {expected_unit}) AS `tmp`
FROM t"""
    assert result == expected


@pytest.mark.parametrize("kind", ["date", "time"])
def test_extract_temporal_from_timestamp(kind):
    t = ibis.table([("ts", dt.timestamp)], name="t")
    expr = getattr(t.ts, kind)()
    result = ibis_bigquery.compile(expr)
    expected = f"""\
SELECT {kind.upper()}(`ts`) AS `tmp`
FROM t"""
    assert result == expected


def test_now():
    expr = ibis.now()
    result = ibis_bigquery.compile(expr)
    expected = "SELECT CURRENT_TIMESTAMP() AS `tmp`"
    assert result == expected


def test_binary():
    t = ibis.table([("value", "double")], name="t")
    expr = t["value"].cast(dt.binary).name("value_hash")
    result = ibis_bigquery.compile(expr)
    expected = """\
SELECT CAST(`value` AS BYTES) AS `tmp`
FROM t"""
    assert result == expected


def test_substring():
    t = ibis.table([("value", "string")], name="t")
    expr = t["value"].substr(3, 1)
    expected = """\
SELECT substr(`value`, 3 + 1, 1) AS `tmp`
FROM t"""
    result = ibis_bigquery.compile(expr)

    assert result == expected


def test_substring_neg_length():
    t = ibis.table([("value", "string")], name="t")
    expr = t["value"].substr(3, -1)
    with pytest.raises(Exception) as exception_info:
        ibis_bigquery.compile(expr)

    expected = "Length parameter should not be a negative value."
    assert str(exception_info.value) == expected


def test_bucket():
    t = ibis.table([("value", "double")], name="t")
    buckets = [0, 1, 3]
    expr = t.value.bucket(buckets).name("foo")
    result = ibis_bigquery.compile(expr)
    expected = """\
SELECT
  CASE
    WHEN (0 <= `value`) AND (`value` < 1) THEN 0
    WHEN (1 <= `value`) AND (`value` <= 3) THEN 1
    ELSE CAST(NULL AS INT64)
  END AS `tmp`
FROM t"""
    expected_2 = """\
SELECT
  CASE
    WHEN (`value` >= 0) AND (`value` < 1) THEN 0
    WHEN (`value` >= 1) AND (`value` <= 3) THEN 1
    ELSE CAST(NULL AS INT64)
  END AS `tmp`
FROM t"""
    assert result == expected or result == expected_2


@pytest.mark.parametrize(
    ("kind", "begin", "end", "expected"),
    [
        ("preceding", None, 1, "UNBOUNDED PRECEDING AND 1 PRECEDING"),
        ("following", 1, None, "1 FOLLOWING AND UNBOUNDED FOLLOWING"),
    ],
)
def test_window_unbounded(kind, begin, end, expected):
    t = ibis.table([("a", "int64")], name="t")
    kwargs = {kind: (begin, end)}
    expr = t.a.sum().over(ibis.window(**kwargs))
    result = ibis_bigquery.compile(expr)
    assert (
        result
        == f"""\
SELECT sum(`a`) OVER (ROWS BETWEEN {expected}) AS `tmp`
FROM t"""
    )


def test_large_compile():
    """
    Tests that compiling a large expression tree finishes
    within a reasonable amount of time
    """
    num_columns = 20
    num_joins = 7

    class MockBackend(ibis_bigquery.Backend):
        pass

    names = [f"col_{i}" for i in range(num_columns)]
    schema = ibis.Schema(names, ["string"] * num_columns)
    ibis_client = MockBackend()
    table = TableExpr(ops.SQLQueryResult("select * from t", schema, ibis_client))
    for _ in range(num_joins):
        table = table.mutate(dummy=ibis.literal(""))
        table = table.left_join(table, ["dummy"])[[table]]

    start = datetime.datetime.now()
    table.compile()
    delta = datetime.datetime.now() - start
    assert delta.total_seconds() < 10



import collections
import functools
import inspect
import itertools
from typing import Dict, Iterable

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz

try:
    from ibis.udf.validate import validate_output_type
except ImportError:
    # 1.2
    def validate_output_type(*args):
        pass


from ibis.expr.signature import Argument as Arg

from ..compiler import BigQueryUDFNode, compiles
from ..datatypes import UDFContext, ibis_type_to_bigquery_type
from ..udf.core import PythonToJavaScriptTranslator

__all__ = ("udf",)

_udf_name_cache: Dict[str, Iterable[int]] = (collections.defaultdict(itertools.count))


def create_udf_node(name, fields):
    """Create a new UDF node type.

    Parameters
    ----------
    name : str
        Then name of the UDF node
    fields : OrderedDict
        Mapping of class member name to definition

    Returns
    -------
    result : type
        A new BigQueryUDFNode subclass
    """
    definition = next(_udf_name_cache[name])
    external_name = "{}_{:d}".format(name, definition)
    return type(external_name, (BigQueryUDFNode,), fields)


def udf(input_type, output_type, strict=True, libraries=None):
    '''Define a UDF for BigQuery

    Parameters
    ----------
    input_type : List[DataType]
    output_type : DataType
    strict : bool
        Whether or not to put a ``'use strict';`` string at the beginning of
        the UDF. Setting to ``False`` is probably a bad idea.
    libraries : List[str]
        A list of Google Cloud Storage URIs containing to JavaScript source
        code. Note that any symbols (functions, classes, variables, etc.) that
        are exposed in these JavaScript files will be visible inside the UDF.

    Returns
    -------
    wrapper : Callable
        The wrapped function

    Notes
    -----
    - ``INT64`` is not supported as an argument type or a return type, as per
      `the BigQuery documentation
      <https://cloud.google.com/bigquery/docs/reference/standard-sql/user-defined-functions#sql-type-encodings-in-javascript>`_.
    - `The follow example doctest doesn't work for Python 3.8
      <https://github.com/ibis-project/ibis/issues/2085>`_.

    Examples
    --------
    >>> if PY38:
    ...     import pytest; pytest.skip("Issue #2085")
    >>> from ibis_bigquery import udf
    >>> import ibis.expr.datatypes as dt
    >>> @udf(input_type=[dt.double], output_type=dt.double)
    ... def add_one(x):
    ...     return x + 1
    >>> print(add_one.js)
    CREATE TEMPORARY FUNCTION add_one_0(x FLOAT64)
    RETURNS FLOAT64
    LANGUAGE js AS """
    'use strict';
    function add_one(x) {
        return (x + 1);
    }
    return add_one(x);
    """;
    >>> @udf(input_type=[dt.double, dt.double],
    ...      output_type=dt.Array(dt.double))
    ... def my_range(start, stop):
    ...     def gen(start, stop):
    ...         curr = start
    ...         while curr < stop:
    ...             yield curr
    ...             curr += 1
    ...     result = []
    ...     for value in gen(start, stop):
    ...         result.append(value)
    ...     return result
    >>> print(my_range.js)
    CREATE TEMPORARY FUNCTION my_range_0(start FLOAT64, stop FLOAT64)
    RETURNS ARRAY<FLOAT64>
    LANGUAGE js AS """
    'use strict';
    function my_range(start, stop) {
        function* gen(start, stop) {
            let curr = start;
            while ((curr < stop)) {
                yield curr;
                curr += 1;
            }
        }
        let result = [];
        for (let value of gen(start, stop)) {
            result.push(value);
        }
        return result;
    }
    return my_range(start, stop);
    """;
    >>> @udf(
    ...     input_type=[dt.double, dt.double],
    ...     output_type=dt.Struct.from_tuples([
    ...         ('width', 'double'), ('height', 'double')
    ...     ])
    ... )
    ... def my_rectangle(width, height):
    ...     class Rectangle:
    ...         def __init__(self, width, height):
    ...             self.width = width
    ...             self.height = height
    ...
    ...         @property
    ...         def area(self):
    ...             return self.width * self.height
    ...
    ...         def perimeter(self):
    ...             return 2 * (self.width + self.height)
    ...
    ...     return Rectangle(width, height)
    >>> print(my_rectangle.js)
    CREATE TEMPORARY FUNCTION my_rectangle_0(width FLOAT64, height FLOAT64)
    RETURNS STRUCT<width FLOAT64, height FLOAT64>
    LANGUAGE js AS """
    'use strict';
    function my_rectangle(width, height) {
        class Rectangle {
            constructor(width, height) {
                this.width = width;
                this.height = height;
            }
            get area() {
                return (this.width * this.height);
            }
            perimeter() {
                return (2 * (this.width + this.height));
            }
        }
        return (new Rectangle(width, height));
    }
    return my_rectangle(width, height);
    """;
    '''
    validate_output_type(output_type)

    if libraries is None:
        libraries = []

    def wrapper(f):
        if not callable(f):
            raise TypeError("f must be callable, got {}".format(f))

        signature = inspect.signature(f)
        parameter_names = signature.parameters.keys()

        udf_node_fields = collections.OrderedDict(
            [
                (name, Arg(rlz.value(type)))
                for name, type in zip(parameter_names, input_type)
            ]
            + [
                (
                    "output_type",
                    lambda self, output_type=output_type: rlz.shape_like(
                        self.args, dtype=output_type
                    ),
                ),
                ("__slots__", ("js",)),
            ]
        )

        udf_node = create_udf_node(f.__name__, udf_node_fields)

        @compiles(udf_node)
        def compiles_udf_node(t, expr):
            return "{}({})".format(
                udf_node.__name__, ", ".join(map(t.translate, expr.op().args))
            )

        type_translation_context = UDFContext()
        return_type = ibis_type_to_bigquery_type(
            dt.dtype(output_type), type_translation_context
        )
        bigquery_signature = ", ".join(
            "{name} {type}".format(
                name=name,
                type=ibis_type_to_bigquery_type(
                    dt.dtype(type), type_translation_context
                ),
            )
            for name, type in zip(parameter_names, input_type)
        )
        source = PythonToJavaScriptTranslator(f).compile()
        js = '''\
CREATE TEMPORARY FUNCTION {external_name}({signature})
RETURNS {return_type}
LANGUAGE js AS """
{strict}{source}
return {internal_name}({args});
"""{libraries};'''.format(
            external_name=udf_node.__name__,
            internal_name=f.__name__,
            return_type=return_type,
            source=source,
            signature=bigquery_signature,
            strict=repr("use strict") + ";\n" if strict else "",
            args=", ".join(parameter_names),
            libraries=(
                "\nOPTIONS (\n    library={}\n)".format(repr(list(libraries)))
                if libraries
                else ""
            ),
        )

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            node = udf_node(*args, **kwargs)
            node.js = js
            return node.to_expr()

        wrapped.__signature__ = signature
        wrapped.js = js
        return wrapped

    return wrapper



import io
import os
import shutil
import tempfile
import urllib.request

import google.auth
import google.auth.exceptions
import ibis  # noqa: F401
import pytest
from google.api_core.exceptions import NotFound
from google.cloud import bigquery

import ibis_bigquery

DEFAULT_PROJECT_ID = "ibis-gbq"
PROJECT_ID_ENV_VAR = "GOOGLE_BIGQUERY_PROJECT_ID"
DATASET_ID = "ibis_gbq_testing"
TESTING_DATA_URI = "https://raw.githubusercontent.com/ibis-project/testing-data/master"

bq = ibis_bigquery.Backend()


def pytest_addoption(parser):
    parser.addoption(
        "--save-dataset",
        action="store_true",
        default=False,
        help="saves all test data in the testing dataset",
    )
    parser.addoption(
        "--no-refresh-dataset",
        action="store_true",
        default=False,
        help="do not refresh the test data in the testing dataset",
    )


@pytest.fixture(scope="session")
def dataset_id() -> str:
    return DATASET_ID


@pytest.fixture(scope="session")
def default_credentials():
    try:
        credentials, project_id = google.auth.default(
            scopes=ibis_bigquery.EXTERNAL_DATA_SCOPES
        )
    except google.auth.exceptions.DefaultCredentialsError as exc:
        pytest.skip(f"Could not get GCP credentials: {exc}")

    return credentials, project_id


@pytest.fixture(scope="session")
def project_id(default_credentials):
    project_id = os.getenv(PROJECT_ID_ENV_VAR)
    if project_id is None:
        _, project_id = default_credentials
    if project_id is None:
        project_id = DEFAULT_PROJECT_ID
    return project_id


@pytest.fixture(scope="session")
def credentials(default_credentials):
    credentials, _ = default_credentials
    return credentials


@pytest.fixture(scope="session")
def client(credentials, project_id):
    return bq.connect(
        project_id=project_id, dataset_id=DATASET_ID, credentials=credentials,
    )


@pytest.fixture(scope="session")
def client2(credentials, project_id):
    return bq.connect(
        project_id=project_id, dataset_id=DATASET_ID, credentials=credentials,
    )


@pytest.fixture(scope="session")
def alltypes(client):
    return client.table("functional_alltypes")


@pytest.fixture(scope="session")
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def parted_alltypes(client):
    return client.table("functional_alltypes_parted")


@pytest.fixture(scope="session")
def parted_df(parted_alltypes):
    return parted_alltypes.execute()


@pytest.fixture(scope="session")
def struct_table(client):
    return client.table("struct_table")


@pytest.fixture(scope="session")
def numeric_table(client):
    return client.table("numeric_table")


@pytest.fixture(scope="session")
def public(project_id, credentials):
    return bq.connect(
        project_id=project_id,
        dataset_id="bigquery-public-data.stackoverflow",
        credentials=credentials,
    )


# Native BigQuery client fixtures
# required to dynamically create the testing dataset,
# the tables, and to populate data into the tables.
@pytest.fixture(scope="session")
def bqclient(client):
    return client.client


# Create testing dataset.
@pytest.fixture(autouse=True, scope="session")
def testing_dataset(bqclient, request, dataset_id):
    dataset_ref = bigquery.DatasetReference(bqclient.project, dataset_id)
    try:
        bqclient.create_dataset(dataset_ref, exists_ok=True)
    except NotFound:
        pass
    yield dataset_ref
    if not request.config.getoption("--save-dataset"):
        bqclient.delete_dataset(dataset_ref, delete_contents=True, not_found_ok=True)


@pytest.fixture(scope="session")
def functional_alltypes_table(testing_dataset):
    return bigquery.TableReference(testing_dataset, "functional_alltypes")


@pytest.fixture(autouse=True, scope="session")
def create_functional_alltypes_table(bqclient, functional_alltypes_table):
    table = bigquery.Table(functional_alltypes_table)
    table.schema = [
        bigquery.SchemaField("index", "INTEGER"),
        bigquery.SchemaField("Unnamed_0", "INTEGER"),
        bigquery.SchemaField("id", "INTEGER"),
        bigquery.SchemaField("bool_col", "BOOLEAN"),
        bigquery.SchemaField("tinyint_col", "INTEGER"),
        bigquery.SchemaField("smallint_col", "INTEGER"),
        bigquery.SchemaField("int_col", "INTEGER"),
        bigquery.SchemaField("bigint_col", "INTEGER"),
        bigquery.SchemaField("float_col", "FLOAT"),
        bigquery.SchemaField("double_col", "FLOAT"),
        bigquery.SchemaField("date_string_col", "STRING"),
        bigquery.SchemaField("string_col", "STRING"),
        bigquery.SchemaField("timestamp_col", "TIMESTAMP"),
        bigquery.SchemaField("year", "INTEGER"),
        bigquery.SchemaField("month", "INTEGER"),
    ]
    bqclient.create_table(table, exists_ok=True)
    return table


@pytest.fixture(autouse=True, scope="session")
def load_functional_alltypes_data(request, bqclient, create_functional_alltypes_table):
    if request.config.getoption("--no-refresh-dataset"):
        return

    table = create_functional_alltypes_table
    load_config = bigquery.LoadJobConfig()
    load_config.skip_leading_rows = 1  # skip the header row.
    load_config.write_disposition = "WRITE_TRUNCATE"
    filepath = download_file("{}/functional_alltypes.csv".format(TESTING_DATA_URI))
    with open(filepath.name, "rb") as csvfile:
        job = bqclient.load_table_from_file(
            csvfile, table, job_config=load_config,
        ).result()
    if job.error_result:
        print("error")


# Ingestion time partitioned table.
@pytest.fixture(scope="session")
def functional_alltypes_parted_table(testing_dataset):
    return bigquery.TableReference(testing_dataset, "functional_alltypes_parted")


@pytest.fixture(scope="session")
def create_functional_alltypes_parted_table(bqclient, functional_alltypes_parted_table):
    table = bigquery.Table(functional_alltypes_parted_table)
    table.schema = [
        bigquery.SchemaField("index", "INTEGER"),
        bigquery.SchemaField("Unnamed_0", "INTEGER"),
        bigquery.SchemaField("id", "INTEGER"),
        bigquery.SchemaField("bool_col", "BOOLEAN"),
        bigquery.SchemaField("tinyint_col", "INTEGER"),
        bigquery.SchemaField("smallint_col", "INTEGER"),
        bigquery.SchemaField("int_col", "INTEGER"),
        bigquery.SchemaField("bigint_col", "INTEGER"),
        bigquery.SchemaField("float_col", "FLOAT"),
        bigquery.SchemaField("double_col", "FLOAT"),
        bigquery.SchemaField("date_string_col", "STRING"),
        bigquery.SchemaField("string_col", "STRING"),
        bigquery.SchemaField("timestamp_col", "TIMESTAMP"),
        bigquery.SchemaField("year", "INTEGER"),
        bigquery.SchemaField("month", "INTEGER"),
    ]
    table.time_partitioning = bigquery.TimePartitioning(
        type_=bigquery.TimePartitioningType.DAY
    )
    table.require_partition_filter = False
    bqclient.create_table(table, exists_ok=True)
    return table


@pytest.fixture(autouse=True, scope="session")
def load_functional_alltypes_parted_data(
    request, bqclient, create_functional_alltypes_parted_table
):
    if request.config.getoption("--no-refresh-dataset"):
        return

    table = create_functional_alltypes_parted_table
    load_config = bigquery.LoadJobConfig()
    load_config.write_disposition = "WRITE_TRUNCATE"
    load_config.skip_leading_rows = 1  # skip the header row.
    filepath = download_file("{}/functional_alltypes.csv".format(TESTING_DATA_URI))
    with open(filepath.name, "rb") as csvfile:
        job = bqclient.load_table_from_file(
            csvfile, table, job_config=load_config,
        ).result()
    if job.error_result:
        print("error")


# Create a table with complex data types (nested and repeated).
@pytest.fixture(scope="session")
def struct_bq_table(testing_dataset):
    return bigquery.TableReference(testing_dataset, "struct_table")


@pytest.fixture(autouse=True, scope="session")
def load_struct_table_data(request, bqclient, struct_bq_table):
    if request.config.getoption("--no-refresh-dataset"):
        return

    load_config = bigquery.LoadJobConfig()
    load_config.write_disposition = "WRITE_TRUNCATE"
    load_config.source_format = "AVRO"
    filepath = download_file("{}/struct_table.avro".format(TESTING_DATA_URI))
    with open(filepath.name, "rb") as avrofile:
        job = bqclient.load_table_from_file(
            avrofile, struct_bq_table, job_config=load_config,
        ).result()
    if job.error_result:
        print("error")


# Create empty date-partitioned table.
@pytest.fixture(scope="session")
def date_table(testing_dataset):
    return bigquery.TableReference(testing_dataset, "date_column_parted")


@pytest.fixture(autouse=True, scope="session")
def create_date_table(bqclient, date_table):
    table = bigquery.Table(date_table)
    table.schema = [
        bigquery.SchemaField("my_date_parted_col", "DATE"),
        bigquery.SchemaField("string_col", "STRING"),
        bigquery.SchemaField("int_col", "INTEGER"),
    ]
    table.time_partitioning = bigquery.TimePartitioning(field="my_date_parted_col")
    bqclient.create_table(table, exists_ok=True)
    return table


# Create empty timestamp-partitioned tables.
@pytest.fixture(scope="session")
def timestamp_table(testing_dataset):
    return bigquery.TableReference(testing_dataset, "timestamp_column_parted")


@pytest.fixture(autouse=True, scope="session")
def create_timestamp_table(bqclient, timestamp_table):
    table = bigquery.Table(timestamp_table)
    table.schema = [
        bigquery.SchemaField("my_timestamp_parted_col", "DATE"),
        bigquery.SchemaField("string_col", "STRING"),
        bigquery.SchemaField("int_col", "INTEGER"),
    ]
    table.time_partitioning = bigquery.TimePartitioning(field="my_timestamp_parted_col")
    bqclient.create_table(table, exists_ok=True)


# Create a table with a numeric column
@pytest.fixture(scope="session")
def numeric_bq_table(testing_dataset):
    return bigquery.TableReference(testing_dataset, "numeric_table")


@pytest.fixture(scope="session")
def create_numeric_table(bqclient, numeric_bq_table):
    table = bigquery.Table(numeric_bq_table)
    table.schema = [
        bigquery.SchemaField("string_col", "STRING"),
        bigquery.SchemaField("numeric_col", "NUMERIC"),
    ]
    bqclient.create_table(table, exists_ok=True)
    return table


@pytest.fixture(autouse=True, scope="session")
def load_numeric_data(request, bqclient, create_numeric_table):
    if request.config.getoption("--no-refresh-dataset"):
        return

    load_config = bigquery.LoadJobConfig()
    load_config.write_disposition = "WRITE_TRUNCATE"
    load_config.source_format = "NEWLINE_DELIMITED_JSON"
    data = """{"string_col": "1st value", "numeric_col": 0.999999999}\n\
               {"string_col": "2nd value", "numeric_col": 0.000000002}"""
    jsonfile = io.StringIO(data)
    table = create_numeric_table
    job = bqclient.load_table_from_file(
        jsonfile, table, job_config=load_config
    ).result()
    if job.error_result:
        print("error")


def download_file(url):
    with urllib.request.urlopen(url) as response:
        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            shutil.copyfileobj(response, tmp_file)
    return tmp_file



"""BigQuery public API."""
import warnings
from typing import Optional, Tuple

import google.auth.credentials
import google.cloud.bigquery as bq
import ibis.expr.schema as sch
import ibis.expr.types as ir
import pydata_google_auth
from google.api_core.exceptions import NotFound
from ibis.backends.base.sql import BaseSQLBackend
from pydata_google_auth import cache

from . import version as ibis_bigquery_version
from .client import (
    BigQueryCursor,
    BigQueryDatabase,
    BigQueryTable,
    _create_client_info,
    bigquery_field_to_ibis_dtype,
    bigquery_param,
    parse_project_and_dataset,
    rename_partitioned_column,
)
from .compiler import BigQueryCompiler

try:
    from .udf import udf  # noqa F401
except ImportError:
    pass


__version__: str = ibis_bigquery_version.__version__

SCOPES = ["https://www.googleapis.com/auth/bigquery"]
EXTERNAL_DATA_SCOPES = [
    "https://www.googleapis.com/auth/bigquery",
    "https://www.googleapis.com/auth/cloud-platform",
    "https://www.googleapis.com/auth/drive",
]
CLIENT_ID = "546535678771-gvffde27nd83kfl6qbrnletqvkdmsese.apps.googleusercontent.com"
CLIENT_SECRET = "iU5ohAF2qcqrujegE3hQ1cPt"


class Backend(BaseSQLBackend):
    name = "bigquery"
    compiler = BigQueryCompiler
    database_class = BigQueryDatabase
    table_class = BigQueryTable

    def connect(
        self,
        project_id: Optional[str] = None,
        dataset_id: str = "",
        credentials: Optional[google.auth.credentials.Credentials] = None,
        application_name: Optional[str] = None,
        auth_local_webserver: bool = True,
        auth_external_data: bool = False,
        auth_cache: str = "default",
        partition_column: Optional[str] = "PARTITIONTIME",
    ) -> "Backend":
        """Create a :class:`Backend` for use with Ibis.

        Parameters
        ----------
        project_id : str
            A BigQuery project id.
        dataset_id : str
            A dataset id that lives inside of the project indicated by
            `project_id`.
        credentials : google.auth.credentials.Credentials
        application_name : str
            A string identifying your application to Google API endpoints.
        auth_local_webserver : bool
            Use a local webserver for the user authentication.  Binds a
            webserver to an open port on localhost between 8080 and 8089,
            inclusive, to receive authentication token. If not set, defaults
            to False, which requests a token via the console.
        auth_external_data : bool
            Authenticate using additional scopes required to `query external
            data sources
            <https://cloud.google.com/bigquery/external-data-sources>`_,
            such as Google Sheets, files in Google Cloud Storage, or files in
            Google Drive. If not set, defaults to False, which requests the
            default BigQuery scopes.
        auth_cache : str
            Selects the behavior of the credentials cache.

            ``'default'``
                Reads credentials from disk if available, otherwise
                authenticates and caches credentials to disk.

            ``'reauth'``
                Authenticates and caches credentials to disk.

            ``'none'``
                Authenticates and does **not** cache credentials.

            Defaults to ``'default'``.
        partition_column : str
            Identifier to use instead of default ``_PARTITIONTIME`` partition
            column. Defaults to ``'PARTITIONTIME'``.

        Returns
        -------
        Backend

        """
        default_project_id = ""

        if credentials is None:
            scopes = SCOPES
            if auth_external_data:
                scopes = EXTERNAL_DATA_SCOPES

            if auth_cache == "default":
                credentials_cache = cache.ReadWriteCredentialsCache(
                    filename="ibis.json"
                )
            elif auth_cache == "reauth":
                credentials_cache = cache.WriteOnlyCredentialsCache(
                    filename="ibis.json"
                )
            elif auth_cache == "none":
                credentials_cache = cache.NOOP
            else:
                raise ValueError(
                    f"Got unexpected value for auth_cache = '{auth_cache}'. "
                    "Expected one of 'default', 'reauth', or 'none'."
                )

            credentials, default_project_id = pydata_google_auth.default(
                scopes,
                client_id=CLIENT_ID,
                client_secret=CLIENT_SECRET,
                credentials_cache=credentials_cache,
                use_local_webserver=auth_local_webserver,
            )

        project_id = project_id or default_project_id

        new_backend = self.__class__()

        (
            new_backend.data_project,
            new_backend.billing_project,
            new_backend.dataset,
        ) = parse_project_and_dataset(project_id, dataset_id)

        new_backend.client = bq.Client(
            project=new_backend.billing_project,
            credentials=credentials,
            client_info=_create_client_info(application_name),
        )
        new_backend.partition_column = partition_column

        return new_backend

    def _parse_project_and_dataset(self, dataset) -> Tuple[str, str]:
        if not dataset and not self.dataset:
            raise ValueError("Unable to determine BigQuery dataset.")
        project, _, dataset = parse_project_and_dataset(
            self.billing_project,
            dataset or "{}.{}".format(self.data_project, self.dataset),
        )
        return project, dataset

    @property
    def project_id(self):
        return self.data_project

    @property
    def dataset_id(self):
        return self.dataset

    def table(self, name, database=None) -> ir.TableExpr:
        t = super().table(name, database=database)
        table_id = self._fully_qualified_name(name, database)
        bq_table = self.client.get_table(table_id)
        return rename_partitioned_column(t, bq_table, self.partition_column)

    def _fully_qualified_name(self, name, database):
        default_project, default_dataset = self._parse_project_and_dataset(database)
        parts = name.split(".")
        if len(parts) == 3:
            return name
        elif len(parts) == 2:
            return "{}.{}".format(default_project, name)
        elif len(parts) == 1:
            return "{}.{}.{}".format(default_project, default_dataset, name)
        raise ValueError("Got too many components in table name: {}".format(name))

    def _get_schema_using_query(self, limited_query):
        with self._execute(limited_query, results=True) as cur:
            # resets the state of the cursor and closes operation
            names, ibis_types = self._adapt_types(cur.description)
        return sch.Schema(names, ibis_types)

    def _get_table_schema(self, qualified_name):
        dataset, table = qualified_name.rsplit(".", 1)
        assert dataset is not None, "dataset is None"
        return self.get_schema(table, database=dataset)

    def _adapt_types(self, descr):
        names = []
        adapted_types = []
        for col in descr:
            names.append(col.name)
            typename = bigquery_field_to_ibis_dtype(col)
            adapted_types.append(typename)
        return names, adapted_types

    def _execute(self, stmt, results=True, query_parameters=None):
        job_config = bq.job.QueryJobConfig()
        job_config.query_parameters = query_parameters or []
        job_config.use_legacy_sql = False  # False by default in >=0.28
        query = self.client.query(
            stmt, job_config=job_config, project=self.billing_project
        )
        query.result()  # blocks until finished
        return BigQueryCursor(query)

    def raw_sql(self, query: str, results=False, params=None):
        query_parameters = [
            bigquery_param(param, value) for param, value in (params or {}).items()
        ]
        return self._execute(query, results=results, query_parameters=query_parameters)

    @property
    def current_database(self) -> str:
        return self.dataset

    def database(self, name=None):
        if name is None and not self.dataset:
            raise ValueError(
                "Unable to determine BigQuery dataset. Call "
                "client.database('my_dataset') or set_database('my_dataset') "
                "to assign your client a dataset."
            )
        return self.database_class(name or self.dataset, self)

    def execute(self, expr, params=None, limit="default", **kwargs):
        """Compile and execute the given Ibis expression.

        Compile and execute Ibis expression using this backend client
        interface, returning results in-memory in the appropriate object type

        Parameters
        ----------
        expr : Expr
        limit : int, default None
          For expressions yielding result yets; retrieve at most this number of
          values/rows. Overrides any limit already set on the expression.
        params : not yet implemented
        kwargs : Backends can receive extra params. For example, clickhouse
            uses this to receive external_tables as dataframes.

        Returns
        -------
        output : input type dependent
          Table expressions: pandas.DataFrame
          Array expressions: pandas.Series
          Scalar expressions: Python scalar value
        """
        # TODO: upstream needs to pass params to raw_sql, I think.
        kwargs.pop("timecontext", None)
        query_ast = self.compiler.to_ast_ensure_limit(expr, limit, params=params)
        sql = query_ast.compile()
        self._log(sql)
        cursor = self.raw_sql(sql, params=params, **kwargs)
        schema = self.ast_schema(query_ast, **kwargs)
        result = self.fetch_from_cursor(cursor, schema)

        if hasattr(getattr(query_ast, "dml", query_ast), "result_handler"):
            result = query_ast.dml.result_handler(result)

        return result

    def exists_database(self, name):
        """
        Return whether a database name exists in the current connection.

        Deprecated in Ibis 2.0. Use `name in client.list_databases()` instead.
        """
        warnings.warn(
            "`client.exists_database(name)` is deprecated, and will be "
            "removed in a future version of Ibis. Use "
            "`name in client.list_databases()` instead.",
            FutureWarning,
        )

        project, dataset = self._parse_project_and_dataset(name)
        client = self.client
        dataset_ref = client.dataset(dataset, project=project)
        try:
            client.get_dataset(dataset_ref)
        except NotFound:
            return False
        else:
            return True

    def exists_table(self, name: str, database: str = None) -> bool:
        """
        Return whether a table name exists in the database.

        Deprecated in Ibis 2.0. Use `name in client.list_tables()` instead.
        """
        warnings.warn(
            "`client.exists_table(name)` is deprecated, and will be "
            "removed in a future version of Ibis. Use "
            "`name in client.list_tables()` instead.",
            FutureWarning,
        )

        table_id = self._fully_qualified_name(name, database)
        client = self.client
        try:
            client.get_table(table_id)
        except NotFound:
            return False
        else:
            return True

    def fetch_from_cursor(self, cursor, schema):
        df = cursor.query.to_dataframe()
        return schema.apply_to(df)

    def get_schema(self, name, database=None):
        table_id = self._fully_qualified_name(name, database)
        table_ref = bq.TableReference.from_string(table_id)
        bq_table = self.client.get_table(table_ref)
        return sch.infer(bq_table)

    def list_databases(self, like=None):
        results = [
            dataset.dataset_id
            for dataset in self.client.list_datasets(project=self.data_project)
        ]
        return self._filter_with_like(results, like)

    def list_tables(self, like=None, database=None):
        project, dataset = self._parse_project_and_dataset(database)
        dataset_ref = bq.DatasetReference(project, dataset)
        result = [table.table_id for table in self.client.list_tables(dataset_ref)]
        return self._filter_with_like(result, like)

    def set_database(self, name):
        self.data_project, self.dataset = self._parse_project_and_dataset(name)

    @property
    def version(self):
        return bq.__version__


def compile(expr, params=None, **kwargs):
    """Compile an expression for BigQuery.
    Returns
    -------
    compiled : str
    See Also
    --------
    ibis.expr.types.Expr.compile
    """
    backend = Backend()
    return backend.compile(expr, params=params, **kwargs)


def connect(
    project_id: Optional[str] = None,
    dataset_id: str = "",
    credentials: Optional[google.auth.credentials.Credentials] = None,
    application_name: Optional[str] = None,
    auth_local_webserver: bool = False,
    auth_external_data: bool = False,
    auth_cache: str = "default",
    partition_column: Optional[str] = "PARTITIONTIME",
) -> Backend:
    """Create a :class:`Backend` for use with Ibis.

    Parameters
    ----------
    project_id : str
        A BigQuery project id.
    dataset_id : str
        A dataset id that lives inside of the project indicated by
        `project_id`.
    credentials : google.auth.credentials.Credentials
    application_name : str
        A string identifying your application to Google API endpoints.
    auth_local_webserver : bool
        Use a local webserver for the user authentication.  Binds a
        webserver to an open port on localhost between 8080 and 8089,
        inclusive, to receive authentication token. If not set, defaults
        to False, which requests a token via the console.
    auth_external_data : bool
        Authenticate using additional scopes required to `query external
        data sources
        <https://cloud.google.com/bigquery/external-data-sources>`_,
        such as Google Sheets, files in Google Cloud Storage, or files in
        Google Drive. If not set, defaults to False, which requests the
        default BigQuery scopes.
    auth_cache : str
        Selects the behavior of the credentials cache.

        ``'default'``
            Reads credentials from disk if available, otherwise
            authenticates and caches credentials to disk.

        ``'reauth'``
            Authenticates and caches credentials to disk.

        ``'none'``
            Authenticates and does **not** cache credentials.

        Defaults to ``'default'``.
    partition_column : str
        Identifier to use instead of default ``_PARTITIONTIME`` partition
        column. Defaults to ``'PARTITIONTIME'``.

    Returns
    -------
    Backend

    """
    backend = Backend()
    return backend.connect(
        project_id=project_id,
        dataset_id=dataset_id,
        credentials=credentials,
        application_name=application_name,
        auth_local_webserver=auth_local_webserver,
        auth_external_data=auth_external_data,
        auth_cache=auth_cache,
        partition_column=partition_column,
    )


__all__ = [
    "__version__",
    "Backend",
    "compile",
    "connect",
]



import ibis  # noqa: F401
import pytest

import ibis_bigquery.client


@pytest.mark.parametrize(
    ["project", "dataset", "expected"],
    [
        ("my-project", "", ("my-project", "my-project", "")),
        ("my-project", "my_dataset", ("my-project", "my-project", "my_dataset"),),
        (
            "billing-project",
            "data-project.my_dataset",
            ("data-project", "billing-project", "my_dataset"),
        ),
    ],
)
def test_parse_project_and_dataset(project, dataset, expected):
    got = ibis_bigquery.client.parse_project_and_dataset(project, dataset)
    assert got == expected


def test_parse_project_and_dataset_raises_error():
    expected_message = "data-project.my_dataset.table is not a BigQuery dataset"
    with pytest.raises(ValueError, match=expected_message):
        ibis_bigquery.client.parse_project_and_dataset(
            "my-project", "data-project.my_dataset.table"
        )



"""Test to verify that the package is discoverable from Ibis v2."""

import ibis


def test_has_bigquery():
    assert hasattr(ibis, "bigquery")


def test_compile():
    expr = ibis.literal(1)
    result = ibis.bigquery.compile(expr)
    expected = "SELECT 1"
    assert expected in result
